import ast
import logging
from _ast import AST

from IPython.terminal.embed import InteractiveShellEmbed
from traitlets.config import Config
from typing_extensions import List, Optional, Tuple, Dict

from .datastructures.enums import PromptFor
from .datastructures.callable_expression import CallableExpression, parse_string_to_expression
from .datastructures.dataclasses import CaseQuery
from .utils import extract_dependencies, contains_return_statement, make_set


class CustomInteractiveShell(InteractiveShellEmbed):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.all_lines = []

    def run_cell(self, raw_cell: str, **kwargs):
        """
        Override the run_cell method to capture return statements.
        """
        if contains_return_statement(raw_cell):
            self.all_lines.append(raw_cell)
            print("Exiting shell on `return` statement.")
            self.history_manager.store_inputs(line_num=self.execution_count, source=raw_cell)
            self.ask_exit()
            return None
        result = super().run_cell(raw_cell, **kwargs)
        if result.error_in_exec is None and result.error_before_exec is None:
            self.all_lines.append(raw_cell)
        return result


class IPythonShell:
    """
    Create an embedded Ipython shell that can be used to prompt the user for input.
    """

    def __init__(self, scope: Optional[Dict] = None, header: Optional[str] = None):
        """
        Initialize the Ipython shell with the given scope and header.

        :param scope: The scope to use for the shell.
        :param header: The header to display when the shell is started.
        """
        self.scope: Dict = scope or {}
        self.header: str = header or ">>> Embedded Ipython Shell"
        self.user_input: Optional[str] = None
        self.shell: CustomInteractiveShell = self._init_shell()
        self.all_code_lines: List[str] = []

    def _init_shell(self):
        """
        Initialize the Ipython shell with a custom configuration.
        """
        cfg = Config()
        shell = CustomInteractiveShell(config=cfg, user_ns=self.scope, banner1=self.header)
        return shell

    def run(self):
        """
        Run the embedded shell.
        """
        while True:
            try:
                self.shell()
                self.update_user_input_from_code_lines()
                break
            except Exception as e:
                logging.error(e)
                print(e)

    def update_user_input_from_code_lines(self):
        """
        Update the user input from the code lines captured in the shell.
        """
        if len(self.shell.all_lines) == 1 and self.shell.all_lines[0].replace('return', '').strip() == '':
            self.user_input = None
        else:
            self.all_code_lines = extract_dependencies(self.shell.all_lines)
            if len(self.all_code_lines) == 1 and self.all_code_lines[0].strip() == '':
                self.user_input = None
            else:
                self.user_input = '\n'.join(self.all_code_lines)


def prompt_user_for_expression(case_query: CaseQuery, prompt_for: PromptFor, prompt_str: Optional[str] = None)\
        -> Tuple[Optional[str], Optional[CallableExpression]]:
    """
    Prompt the user for an executable python expression to the given case query.

    :param case_query: The case query to prompt the user for.
    :param prompt_for: The type of information ask user about.
    :param prompt_str: The prompt string to display to the user.
    :return: A callable expression that takes a case and executes user expression on it.
    """
    while True:
        user_input, expression_tree = prompt_user_about_case(case_query, prompt_for, prompt_str)
        if user_input is None:
            if prompt_for == PromptFor.Conclusion:
                print("No conclusion provided. Exiting.")
                return None, None
            else:
                print("Conditions must be provided. Please try again.")
                continue
        conclusion_type = bool if prompt_for == PromptFor.Conditions else case_query.attribute_type
        callable_expression = CallableExpression(user_input, conclusion_type, expression_tree=expression_tree,
                                                 scope=case_query.scope)
        try:
            callable_expression(case_query.case)
            break
        except Exception as e:
            logging.error(e)
            print(e)
    return user_input, callable_expression


def prompt_user_about_case(case_query: CaseQuery, prompt_for: PromptFor,
                           prompt_str: Optional[str] = None) -> Tuple[Optional[str], Optional[AST]]:
    """
    Prompt the user for input.

    :param case_query: The case query to prompt the user for.
    :param prompt_for: The type of information the user should provide for the given case.
    :param prompt_str: The prompt string to display to the user.
    :return: The user input, and the executable expression that was parsed from the user input.
    """
    if prompt_str is None:
        prompt_str = f"Give {prompt_for} for {case_query.name}"
    scope = {'case': case_query.case, **case_query.scope}
    shell = IPythonShell(scope=scope, header=prompt_str)
    return prompt_user_input_and_parse_to_expression(shell=shell)


def prompt_user_input_and_parse_to_expression(shell: Optional[IPythonShell] = None,
                                              user_input: Optional[str] = None)\
        -> Tuple[Optional[str], Optional[ast.AST]]:
    """
    Prompt the user for input.

    :param shell: The Ipython shell to use for prompting the user.
    :param user_input: The user input to use. If given, the user input will be used instead of prompting the user.
    :return: The user input and the AST tree.
    """
    while True:
        if user_input is None:
            shell = IPythonShell() if shell is None else shell
            shell.run()
            user_input = shell.user_input
            if user_input is None:
                return None, None
            print(user_input)
        try:
            return user_input, parse_string_to_expression(user_input)
        except Exception as e:
            msg = f"Error parsing expression: {e}"
            logging.error(msg)
            print(msg)
            user_input = None



import os
from unittest import TestCase

import sqlalchemy.orm
from sqlalchemy import select
from sqlalchemy.orm import MappedColumn as Column
from typing_extensions import List, Sequence
import pandas as pd

from ripple_down_rules.datasets import Base, Animal, Species, get_dataset, Habitat, HabitatTable
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR, MultiClassRDR, GeneralRDR
from ripple_down_rules.utils import render_tree, make_set
from test_helpers.helpers import get_fit_scrdr


class TestAlchemyRDR(TestCase):
    session: sqlalchemy.orm.Session
    test_results_dir: str = "./test_results"
    expert_answers_dir: str = "./test_expert_answers"
    cache_file: str = f"{test_results_dir}/zoo_dataset.pkl"
    all_cases: Sequence[Animal]
    targets: List[Species]

    @classmethod
    def setUpClass(cls):
        # load dataset
        zoo = get_dataset(111, cls.cache_file)

        # get data and targets (as pandas dataframes)
        X = zoo['features']
        y = zoo['targets']
        names = zoo['ids'].values.flatten()
        X.loc[:, "name"] = names

        cls._init_session_and_insert_data(X)

        # init cases
        query = select(Animal)
        cls.all_cases = cls.session.scalars(query).all()

        # init targets
        category_names = ["mammal", "bird", "reptile", "fish", "amphibian", "insect", "molusc"]
        category_id_to_name = {i + 1: name for i, name in enumerate(category_names)}
        cls.targets = [Species(category_id_to_name[i]) for i in y.values.flatten()]

    @classmethod
    def _init_session_and_insert_data(cls, data: pd.DataFrame):
        engine = sqlalchemy.create_engine("sqlite:///:memory:")
        Base.metadata.create_all(engine)
        session = sqlalchemy.orm.Session(engine)
        session.bulk_insert_mappings(Animal, data.to_dict(orient="records"))
        session.commit()
        cls.session = session

    def test_fit_scrdr(self):
        use_loaded_answers = True
        draw_tree = False
        filename = self.expert_answers_dir + "/scrdr_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        query = select(Animal)
        result = self.session.scalars(query).all()
        scrdr = SingleClassRDR()
        case_queries = [CaseQuery(c, "species", (Species,), True, _target=t)
                        for c, t in zip(self.all_cases, self.targets)]
        scrdr.fit(case_queries, expert=expert,
                  animate_tree=draw_tree, session=self.session)

        cat = scrdr.classify(result[50])
        assert cat == self.targets[50]

    def test_fit_mcrdr_stop_only(self):
        use_loaded_answers = True
        draw_tree = False
        expert, filename = self.get_expert_and_file_name(use_loaded_answers,
                                                         "mcrdr_expert_answers_stop_only_fit")

        mcrdr = MultiClassRDR()
        case_queries = [CaseQuery(c, "species", (Species,), True, _target=t)
                        for c, t in zip(self.all_cases, self.targets)]
        mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)

        for case_query in case_queries:
            cat = mcrdr.classify(case_query.case)
            assert make_set(cat) == make_set(case_query.target_value)

    def test_fit_grdr(self):
        use_loaded_answers = True
        save_answers = False
        draw_tree = False
        filename = self.expert_answers_dir + "/grdr_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        fit_scrdr, _ = get_fit_scrdr(self.all_cases, self.targets)

        grdr = GeneralRDR()
        grdr.add_rdr(fit_scrdr)

        def get_habitat(x: Animal, t: Column) -> List[Column]:
            habitats = set()
            if t == Species.mammal and x.aquatic == 0:
                habitats = {HabitatTable(Habitat.land)}
            elif t == Species.bird:
                habitats = {HabitatTable(Habitat.land)}
                if x.airborne == 1:
                    habitats.update({HabitatTable(Habitat.air)})
                if x.aquatic == 1:
                    habitats.update({HabitatTable(Habitat.water)})
            elif t == Species.fish:
                habitats = {HabitatTable(Habitat.water)}
            elif t == Species.molusc:
                habitats = {HabitatTable(Habitat.land)}
                if x.aquatic == 1:
                    habitats.update({HabitatTable(Habitat.water)})
            if len(habitats) == 0:
                return {'species': t}
            else:
                return {'habitats': habitats, 'species': t}

        n = 20
        habitat_targets = [get_habitat(x, t) for x, t in zip(self.all_cases[:n], self.targets[:n])]
        case_queries = []
        for case, targets in zip(self.all_cases[:n], habitat_targets):
            for attr, target in targets.items():
                case_queries.append(CaseQuery(case, attr, (Species,) if attr == 'species' else (Habitat,),
                                              True if attr == 'species' else False, _target=target))
        grdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        for rule in grdr.start_rules:
            render_tree(rule, use_dot_exporter=True,
                        filename=self.test_results_dir + f"/grdr_{type(rule.conclusion).__name__}")

        for case, case_targets in zip(self.all_cases[:n], habitat_targets):
            cat = grdr.classify(case)
            for cat_name, cat_val in cat.items():
                assert make_set(cat_val) == make_set(case_targets[cat_name])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def get_expert_and_file_name(self, use_loaded_answers: bool, filename: str):
        filename = self.expert_answers_dir + "/" + filename
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)
        return expert, filename


# tests = TestAlchemyRDR()
# tests.setUpClass()
# tests.test_fit_scrdr()
# tests.test_fit_mcrdr_stop_only()
# tests.test_fit_grdr()



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from enum import Enum

from anytree import NodeMixin
from typing_extensions import List, Optional, Self, Union, Dict, Any, Tuple

from .datastructures.callable_expression import CallableExpression
from .datastructures.case import Case
from sqlalchemy.orm import DeclarativeBase as SQLTable
from .datastructures.enums import RDREdge, Stop
from .utils import SubclassJSONSerializer, is_iterable, get_full_class_name, conclusion_to_json, \
    get_rule_conclusion_as_source_code


class Rule(NodeMixin, SubclassJSONSerializer, ABC):
    fired: Optional[bool] = None
    """
    Whether the rule has fired or not.
    """

    def __init__(self, conditions: Optional[CallableExpression] = None,
                 conclusion: Optional[CallableExpression] = None,
                 parent: Optional[Rule] = None,
                 corner_case: Optional[Union[Case, SQLTable]] = None,
                 weight: Optional[str] = None,
                 conclusion_name: Optional[str] = None):
        """
        A rule in the ripple down rules classifier.

        :param conditions: The conditions of the rule.
        :param conclusion: The conclusion of the rule when the conditions are met.
        :param parent: The parent rule of this rule.
        :param corner_case: The corner case that this rule is based on/created from.
        :param weight: The weight of the rule, which is the type of edge connecting the rule to its parent.
        :param conclusion_name: The name of the conclusion of the rule.
        """
        super(Rule, self).__init__()
        self.conclusion = conclusion
        self.corner_case = corner_case
        self.parent = parent
        self.weight: Optional[str] = weight
        self.conditions = conditions if conditions else None
        self.conclusion_name: Optional[str] = conclusion_name
        self.json_serialization: Optional[Dict[str, Any]] = None
        self._name: Optional[str] = None

    def _post_detach(self, parent):
        """
        Called after this node is detached from the tree, useful when drawing the tree.

        :param parent: The parent node from which this node was detached.
        """
        self.weight = None

    def __call__(self, x: Case) -> Self:
        return self.evaluate(x)

    def evaluate(self, x: Case) -> Rule:
        """
        Check if the rule or its refinement or its alternative match the case,
        by checking if the conditions are met, then return the rule that matches the case.

        :param x: The case to evaluate the rule on.
        :return: The rule that fired or the last evaluated rule if no rule fired.
        """
        if not self.conditions:
            raise ValueError("Rule has no conditions")
        self.fired = self.conditions(x)
        return self.evaluate_next_rule(x)

    @abstractmethod
    def evaluate_next_rule(self, x: Case):
        """
        Evaluate the next rule after this rule is evaluated.
        """
        pass

    def write_conclusion_as_source_code(self, parent_indent: str = "", defs_file: Optional[str] = None) -> str:
        """
        Get the source code representation of the conclusion of the rule.

        :param parent_indent: The indentation of the parent rule.
        :param defs_file: The file to write the conclusion to if it is a definition.
        :return: The source code representation of the conclusion of the rule.
        """
        if self.conclusion.user_input is not None:
            conclusion = self.conclusion.user_input
        else:
            conclusion = self.conclusion.conclusion
        conclusion_func, conclusion_func_call = self._conclusion_source_code(conclusion, parent_indent=parent_indent)
        if conclusion_func is not None:
            with open(defs_file, 'a') as f:
                f.write(conclusion_func.strip() + "\n\n\n")
        return conclusion_func_call

    @abstractmethod
    def _conclusion_source_code(self, conclusion: Any, parent_indent: str = "") -> Tuple[Optional[str], str]:
        pass

    def write_condition_as_source_code(self, parent_indent: str = "", defs_file: Optional[str] = None) -> str:
        """
        Get the source code representation of the conditions of the rule.

        :param parent_indent: The indentation of the parent rule.
        :param defs_file: The file to write the conditions to if they are a definition.
        """
        if_clause = self._if_statement_source_code_clause()
        if '\n' not in self.conditions.user_input:
            return f"{parent_indent}{if_clause} {self.conditions.user_input}:\n"
        elif "def " in self.conditions.user_input:
            if defs_file is None:
                raise ValueError("Cannot write conditions to source code as definitions python file was not given.")
            # This means the conditions are a definition that should be written and then called
            conditions_lines = self.conditions.user_input.split('\n')
            # use regex to replace the function name
            new_function_name = f"def conditions_{id(self)}"
            conditions_lines[0] = re.sub(r"def (\w+)", new_function_name, conditions_lines[0])
            def_code = "\n".join(conditions_lines)
            with open(defs_file, 'a') as f:
                f.write(def_code.strip() + "\n\n\n")
            return f"\n{parent_indent}{if_clause} {new_function_name.replace('def ', '')}(case):\n"

    @abstractmethod
    def _if_statement_source_code_clause(self) -> str:
        pass

    def _to_json(self) -> Dict[str, Any]:
        json_serialization = {"conditions": self.conditions.to_json(),
                              "conclusion": conclusion_to_json(self.conclusion),
                              "parent": self.parent.json_serialization if self.parent else None,
                              "corner_case": SubclassJSONSerializer.to_json_static(self.corner_case),
                              "conclusion_name": self.conclusion_name,
                              "weight": self.weight}
        return json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Rule:
        loaded_rule = cls(conditions=CallableExpression.from_json(data["conditions"]),
                          conclusion=CallableExpression.from_json(data["conclusion"]),
                          parent=cls.from_json(data["parent"]),
                          corner_case=Case.from_json(data["corner_case"]),
                          conclusion_name=data["conclusion_name"],
                          weight=data["weight"])
        return loaded_rule

    @property
    def name(self):
        """
        Get the name of the rule, which is the conditions and the conclusion.
        """
        return self._name if self._name is not None else self.__str__()

    @name.setter
    def name(self, new_name: str):
        """
        Set the name of the rule.
        """
        self._name = new_name

    def __str__(self, sep="\n"):
        """
        Get the string representation of the rule, which is the conditions and the conclusion.
        """
        return f"{self.conditions}{sep}=> {self.conclusion}"

    def __repr__(self):
        return self.__str__()


class HasAlternativeRule:
    """
    A mixin class for rules that have an alternative rule.
    """
    _alternative: Optional[Rule] = None
    """
    The alternative rule of the rule, which is evaluated when the rule doesn't fire.
    """
    furthest_alternative: Optional[List[Rule]] = None
    """
    The furthest alternative rule of the rule, which is the last alternative rule in the chain of alternative rules.
    """
    all_alternatives: Optional[List[Rule]] = None
    """
    All alternative rules of the rule, which is all the alternative rules in the chain of alternative rules.
    """

    @property
    def alternative(self) -> Optional[Rule]:
        return self._alternative

    @alternative.setter
    def alternative(self, new_rule: Rule):
        """
        Set the alternative rule of the rule. It is important that no rules should be retracted or changed,
        only new rules should be added.
        """
        if new_rule is None:
            return
        if self.furthest_alternative:
            self.furthest_alternative[-1].alternative = new_rule
        else:
            new_rule.parent = self
            new_rule.weight = RDREdge.Alternative.value if not new_rule.weight else new_rule.weight
            self._alternative = new_rule
        self.furthest_alternative = [new_rule]


class HasRefinementRule:
    _refinement: Optional[HasAlternativeRule] = None
    """
    The refinement rule of the rule, which is evaluated when the rule fires.
    """

    @property
    def refinement(self) -> Optional[Rule]:
        return self._refinement

    @refinement.setter
    def refinement(self, new_rule: Rule):
        """
        Set the refinement rule of the rule. It is important that no rules should be retracted or changed,
        only new rules should be added.
        """
        if new_rule is None:
            return
        new_rule.top_rule = self
        if self.refinement:
            self.refinement.alternative = new_rule
        else:
            new_rule.parent = self
            new_rule.weight = RDREdge.Refinement.value
            self._refinement = new_rule


class SingleClassRule(Rule, HasAlternativeRule, HasRefinementRule):
    """
    A rule in the SingleClassRDR classifier, it can have a refinement or an alternative rule or both.
    """

    def evaluate_next_rule(self, x: Case) -> SingleClassRule:
        if self.fired:
            returned_rule = self.refinement(x) if self.refinement else self
        else:
            returned_rule = self.alternative(x) if self.alternative else self
        return returned_rule if returned_rule.fired else self

    def fit_rule(self, x: Case, target: CallableExpression, conditions: CallableExpression):
        new_rule = SingleClassRule(conditions, target,
                                   corner_case=x, parent=self)
        if self.fired:
            self.refinement = new_rule
        else:
            self.alternative = new_rule

    def _to_json(self) -> Dict[str, Any]:
        self.json_serialization = {**super(SingleClassRule, self)._to_json(),
                                   "refinement": self.refinement.to_json() if self.refinement is not None else None,
                                   "alternative": self.alternative.to_json() if self.alternative is not None else None}
        return self.json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> SingleClassRule:
        loaded_rule = super(SingleClassRule, cls)._from_json(data)
        loaded_rule.refinement = SingleClassRule.from_json(data["refinement"])
        loaded_rule.alternative = SingleClassRule.from_json(data["alternative"])
        return loaded_rule

    def _conclusion_source_code(self, conclusion: Any, parent_indent: str = "") -> Tuple[Optional[str], str]:
        conclusion = str(conclusion)
        # indent = parent_indent + " " * 4
        # if '\n' not in conclusion:
        #     return None, f"{indent}return {conclusion}\n"
        # else:
        return get_rule_conclusion_as_source_code(self, conclusion, parent_indent=parent_indent)

    def _if_statement_source_code_clause(self) -> str:
        return "elif" if self.weight == RDREdge.Alternative.value else "if"


class MultiClassStopRule(Rule, HasAlternativeRule):
    """
    A rule in the MultiClassRDR classifier, it can have an alternative rule and a top rule,
    the conclusion of the rule is a Stop category meant to stop the parent conclusion from being made.
    """
    top_rule: Optional[MultiClassTopRule] = None
    """
    The top rule of the rule, which is the nearest ancestor that fired and this rule is a refinement of.
    """

    def __init__(self, *args, **kwargs):
        super(MultiClassStopRule, self).__init__(*args, **kwargs)
        self.conclusion = CallableExpression(conclusion_type=(Stop,), conclusion=Stop.stop)

    def evaluate_next_rule(self, x: Case) -> Optional[Union[MultiClassStopRule, MultiClassTopRule]]:
        if self.fired:
            self.top_rule.fired = False
            return self.top_rule.alternative
        elif self.alternative:
            return self.alternative(x)
        else:
            return self.top_rule.alternative

    def _to_json(self) -> Dict[str, Any]:
        self.json_serialization = {**Rule._to_json(self),
                                   "alternative": self.alternative.to_json() if self.alternative is not None else None}
        return self.json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> MultiClassStopRule:
        loaded_rule = super(MultiClassStopRule, cls)._from_json(data)
        # The following is done to prevent re-initialization of the top rule,
        # so the top rule that is already initialized is passed in the data instead of its json serialization.
        loaded_rule.top_rule = data['top_rule']
        if data['alternative'] is not None:
            data['alternative']['top_rule'] = data['top_rule']
        loaded_rule.alternative = MultiClassStopRule.from_json(data["alternative"])
        return loaded_rule

    def _conclusion_source_code(self, conclusion: Any, parent_indent: str = "") -> Tuple[None, str]:
        return None, f"{parent_indent}{' ' * 4}pass\n"

    def _if_statement_source_code_clause(self) -> str:
        return "elif" if self.weight == RDREdge.Alternative.value else "if"


class MultiClassTopRule(Rule, HasRefinementRule, HasAlternativeRule):
    """
    A rule in the MultiClassRDR classifier, it can have a refinement and a next rule.
    """

    def __init__(self, *args, **kwargs):
        super(MultiClassTopRule, self).__init__(*args, **kwargs)
        self.weight = RDREdge.Next.value

    def evaluate_next_rule(self, x: Case) -> Optional[Union[MultiClassStopRule, MultiClassTopRule]]:
        if self.fired and self.refinement:
            return self.refinement(x)
        elif self.alternative:  # Here alternative refers to next rule in MultiClassRDR
            return self.alternative

    def fit_rule(self, x: Case, target: CallableExpression, conditions: CallableExpression):
        if self.fired and target != self.conclusion:
            self.refinement = MultiClassStopRule(conditions, corner_case=x, parent=self)
        elif not self.fired:
            self.alternative = MultiClassTopRule(conditions, target, corner_case=x, parent=self)

    def _to_json(self) -> Dict[str, Any]:
        self.json_serialization = {**Rule._to_json(self),
                                   "refinement": self.refinement.to_json() if self.refinement is not None else None,
                                   "alternative": self.alternative.to_json() if self.alternative is not None else None}
        return self.json_serialization

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> MultiClassTopRule:
        loaded_rule = super(MultiClassTopRule, cls)._from_json(data)
        # The following is done to prevent re-initialization of the top rule,
        # so the top rule that is already initialized is passed in the data instead of its json serialization.
        if data['refinement'] is not None:
            data['refinement']['top_rule'] = loaded_rule
        loaded_rule.refinement = MultiClassStopRule.from_json(data["refinement"])
        loaded_rule.alternative = MultiClassTopRule.from_json(data["alternative"])
        return loaded_rule

    def _conclusion_source_code(self, conclusion: Any, parent_indent: str = "") -> Tuple[str, str]:
        conclusion_str = str(conclusion)
        indent = parent_indent + " " * 4
        # if '\n' not in conclusion_str:
        #     func = None
        #     if is_iterable(conclusion):
        #         conclusion_str = "{" + ", ".join([str(c) for c in conclusion]) + "}"
        #     else:
        #         conclusion_str = "{" + str(conclusion) + "}"
        # else:
        func, func_call = get_rule_conclusion_as_source_code(self, conclusion_str, parent_indent=parent_indent)
        conclusion_str = func_call.replace("return ", "").strip()

        statement = f"{indent}conclusions.update(make_set({conclusion_str}))\n"
        if self.alternative is None:
            statement += f"{parent_indent}return conclusions\n"
        return func, statement

    def _if_statement_source_code_clause(self) -> str:
        return "if"



"""
This file contains decorators for the RDR (Ripple Down Rules) framework. Where each type of RDR has a decorator
that can be used with any python function such that this function can benefit from the incremental knowledge acquisition
of the RDRs.
"""
import os.path
from functools import wraps
from typing import Callable, Optional, Type

from sqlalchemy.orm import Session
from typing_extensions import Any

from ripple_down_rules.datastructures import Case, Category, create_case, CaseQuery
from ripple_down_rules.experts import Expert, Human

from ripple_down_rules.rdr import SingleClassRDR, MultiClassRDR, GeneralRDR
from ripple_down_rules.utils import get_method_args_as_dict, get_method_name, get_method_class_name_if_exists, \
    get_method_file_name, get_func_rdr_model_path


def single_class_rdr(
    model_dir: str,
    fit: bool = True,
    expert: Optional[Expert] = None,
    session: Optional[Session] = None,
) -> Callable:
    """
    Decorator to use a SingleClassRDR as a classifier.
    """
    expert = expert if expert else Human(session=session)

    def decorator(func: Callable) -> Callable:
        scrdr_model_path = get_func_rdr_model_path(func, model_dir)
        if os.path.exists(scrdr_model_path):
            scrdr = SingleClassRDR.load(scrdr_model_path)
            scrdr.session = session
        else:
            scrdr = SingleClassRDR(session=session)

        @wraps(func)
        def wrapper(*args, **kwargs) -> Category:
            case_dict = get_method_args_as_dict(func, *args, **kwargs)
            func_output = func(*args, **kwargs)
            if func_output is not None:
                case_dict.update({"_output": func_output})
            case = create_case(case_dict, recursion_idx=3)
            if fit:
                output = scrdr.fit_case(CaseQuery(case), expert=expert)
                scrdr.save(scrdr_model_path)
                return output
            else:
                return scrdr.classify(case)
        return wrapper

    return decorator



from __future__ import annotations

from enum import auto, Enum

from typing_extensions import List, Dict, Any

from ripple_down_rules.utils import SubclassJSONSerializer


class Category(str, SubclassJSONSerializer, Enum):

    @classmethod
    def from_str(cls, value: str) -> Category:
        return getattr(cls, value)

    @classmethod
    def from_strs(cls, values: List[str]) -> List[Category]:
        return [cls.from_str(value) for value in values]

    @property
    def as_dict(self):
        return {self.__class__.__name__.lower(): self.value}

    def _to_json(self) -> Dict[str, Any]:
        return self.as_dict

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Category:
        return cls.from_str(data[cls.__name__.lower()])


class Stop(Category):
    """
    A stop category is a special category that represents the stopping of the classification to prevent a wrong
    conclusion from being made.
    """
    stop = "stop"


class ExpressionParser(Enum):
    """
    Parsers for expressions to evaluate and encapsulate the expression into a callable function.
    """
    ASTVisitor: int = auto()
    """
    Generic python Abstract Syntax Tree that detects variables, attributes, binary/boolean expressions , ...etc.
    """
    SQLAlchemy: int = auto()
    """
    Specific for SQLAlchemy expressions on ORM Tables.
    """


class PromptFor(Enum):
    """
    The reason of the prompt. (e.g. get conditions, conclusions, or affirmation).
    """
    Conditions: str = "conditions"
    """
    Prompt for rule conditions about a case.
    """
    Conclusion: str = "conclusion"
    """
    Prompt for rule conclusion about a case.
    """
    Affirmation: str = "affirmation"
    """
    Prompt for rule conclusion affirmation about a case.
    """

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()


class CategoricalValue(Enum):
    """
    A categorical value is a value that is a category.
    """

    def __eq__(self, other):
        if isinstance(other, CategoricalValue):
            return self.name == other.name
        elif isinstance(other, str):
            return self.name == other
        return self.name == other

    def __hash__(self):
        return hash(self.name)

    @classmethod
    def to_list(cls):
        return list(cls._value2member_map_.keys())

    @classmethod
    def from_str(cls, category: str):
        return cls[category.lower()]

    @classmethod
    def from_strs(cls, categories: List[str]):
        return [cls.from_str(c) for c in categories]

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()


class RDRMode(Enum):
    Propositional = auto()
    """
    Propositional mode, the mode where the rules are propositional.
    """
    Relational = auto()
    """
    Relational mode, the mode where the rules are relational.
    """


class MCRDRMode(Enum):
    """
    The modes of the MultiClassRDR.
    """
    StopOnly = auto()
    """
    StopOnly mode, stop wrong conclusion from being made and does not add a new rule to make the correct conclusion.
    """
    StopPlusRule = auto()
    """
    StopPlusRule mode, stop wrong conclusion from being made and adds a new rule with same conditions as stopping rule
     to make the correct conclusion.
    """
    StopPlusRuleCombined = auto()
    """
    StopPlusRuleCombined mode, stop wrong conclusion from being made and adds a new rule with combined conditions of
    stopping rule and the rule that should have fired.
    """


class RDREdge(Enum):
    Refinement = "except if"
    """
    Refinement edge, the edge that represents the refinement of an incorrectly fired rule.
    """
    Alternative = "else if"
    """
    Alternative edge, the edge that represents the alternative to the rule that has not fired.
    """
    Next = "next"
    """
    Next edge, the edge that represents the next rule to be evaluated.
    """


class ValueType(Enum):
    Unary = auto()
    """
    Unary value type (eg. null).
    """
    Binary = auto()
    """
    Binary value type (eg. True, False).
    """
    Discrete = auto()
    """
    Discrete value type (eg. 1, 2, 3).
    """
    Continuous = auto()
    """
    Continuous value type (eg. 1.0, 2.5, 3.4).
    """
    Nominal = auto()
    """
    Nominal value type (eg. red, blue, green), categories where the values have no natural order.
    """
    Ordinal = auto()
    """
    Ordinal value type (eg. low, medium, high), categories where the values have a natural order.
    """
    Iterable = auto()
    """
    Iterable value type (eg. [1, 2, 3]).
    """



from __future__ import annotations

import importlib
import sys
from abc import ABC, abstractmethod
from copy import copy
from dataclasses import is_dataclass
from io import TextIOWrapper
from types import ModuleType

from matplotlib import pyplot as plt
from ordered_set import OrderedSet
from sqlalchemy.orm import DeclarativeBase as SQLTable
from typing_extensions import List, Optional, Dict, Type, Union, Any, Self, Tuple, Callable, Set

from .datastructures.callable_expression import CallableExpression
from .datastructures.case import Case, CaseAttribute, create_case
from .datastructures.dataclasses import CaseQuery
from .datastructures.enums import MCRDRMode, PromptFor
from .experts import Expert, Human
from .helpers import is_matching
from .rules import Rule, SingleClassRule, MultiClassTopRule, MultiClassStopRule
from .utils import draw_tree, make_set, copy_case, \
    SubclassJSONSerializer, is_iterable, make_list, get_type_from_string, \
    get_case_attribute_type, is_conflicting, update_case


class RippleDownRules(SubclassJSONSerializer, ABC):
    """
    The abstract base class for the ripple down rules classifiers.
    """
    fig: Optional[plt.Figure] = None
    """
    The figure to draw the tree on.
    """
    expert_accepted_conclusions: Optional[List[CaseAttribute]] = None
    """
    The conclusions that the expert has accepted, such that they are not asked again.
    """
    _generated_python_file_name: Optional[str] = None
    """
    The name of the generated python file.
    """

    def __init__(self, start_rule: Optional[Rule] = None):
        """
        :param start_rule: The starting rule for the classifier.
        """
        self.start_rule = start_rule
        self.fig: Optional[plt.Figure] = None

    def fit(self, case_queries: List[CaseQuery],
            expert: Optional[Expert] = None,
            n_iter: int = None,
            animate_tree: bool = False,
            **kwargs_for_fit_case):
        """
        Fit the classifier to a batch of cases and categories.

        :param case_queries: The cases and categories to fit the classifier to.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param n_iter: The number of iterations to fit the classifier for.
        :param animate_tree: Whether to draw the tree while fitting the classifier.
        :param kwargs_for_fit_case: The keyword arguments to pass to the fit_case method.
        """
        targets = []
        if animate_tree:
            plt.ion()
        i = 0
        stop_iterating = False
        num_rules: int = 0
        while not stop_iterating:
            for case_query in case_queries:
                pred_cat = self.fit_case(case_query, expert=expert, **kwargs_for_fit_case)
                if case_query.target is None:
                    continue
                target = {case_query.attribute_name: case_query.target(case_query.case)}
                if len(targets) < len(case_queries):
                    targets.append(target)
                match = is_matching(self.classify, case_query, pred_cat)
                if not match:
                    print(f"Predicted: {pred_cat} but expected: {target}")
                if animate_tree and self.start_rule.size > num_rules:
                    num_rules = self.start_rule.size
                    self.update_figures()
            i += 1
            all_predictions = [1 if is_matching(self.classify, case_query) else 0 for case_query in case_queries
                               if case_query.target is not None]
            all_pred = sum(all_predictions)
            print(f"Accuracy: {all_pred}/{len(targets)}")
            all_predicted = targets and all_pred == len(targets)
            num_iter_reached = n_iter and i >= n_iter
            stop_iterating = all_predicted or num_iter_reached
            if stop_iterating:
                break
        print(f"Finished training in {i} iterations")
        if animate_tree:
            plt.ioff()
            plt.show()

    def __call__(self, case: Union[Case, SQLTable]) -> CaseAttribute:
        return self.classify(case)

    @abstractmethod
    def classify(self, case: Union[Case, SQLTable]) -> Optional[CaseAttribute]:
        """
        Classify a case.

        :param case: The case to classify.
        :return: The category that the case belongs to.
        """
        pass

    def fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CaseAttribute, CallableExpression]:
        """
        Fit the classifier to a case and ask the expert for refinements or alternatives if the classification is
        incorrect by comparing the case with the target category.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        if case_query is None:
            raise ValueError("The case query cannot be None.")
        if case_query.target is None:
            expert.ask_for_conclusion(case_query)
            if case_query.target is None:
                return self.classify(case_query.case)

        self.update_start_rule(case_query, expert)

        return self._fit_case(case_query, expert=expert, **kwargs)

    @abstractmethod
    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CaseAttribute, CallableExpression]:
        """
        Fit the RDR on a case, and ask the expert for refinements or alternatives if the classification is incorrect by
        comparing the case with the target category.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        pass

    @abstractmethod
    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        pass

    def update_figures(self):
        """
        Update the figures of the classifier.
        """
        if isinstance(self, GeneralRDR):
            for i, (rdr_name, rdr) in enumerate(self.start_rules_dict.items()):
                if not rdr.fig:
                    rdr.fig = plt.figure(f"Rule {i}: {rdr_name}")
                draw_tree(rdr.start_rule, rdr.fig)
        else:
            if not self.fig:
                self.fig = plt.figure(0)
            draw_tree(self.start_rule, self.fig)

    @staticmethod
    def case_has_conclusion(case: Union[Case, SQLTable], conclusion_name: str) -> bool:
        """
        Check if the case has a conclusion.

        :param case: The case to check.
        :param conclusion_name: The target category name to compare the case with.
        :return: Whether the case has a conclusion or not.
        """
        return hasattr(case, conclusion_name) and getattr(case, conclusion_name) is not None

    @property
    def type_(self):
        return self.__class__


class RDRWithCodeWriter(RippleDownRules, ABC):

    @abstractmethod
    def write_rules_as_source_code_to_file(self, rule: Rule, file, parent_indent: str = "",
                                           defs_file: Optional[str] = None):
        """
        Write the rules as source code to a file.

        :param rule: The rule to write as source code.
        :param file: The file to write the source code to.
        :param parent_indent: The indentation of the parent rule.
        :param defs_file: The file to write the definitions to.
        """
        pass

    def write_to_python_file(self, file_path: str, postfix: str = ""):
        """
        Write the tree of rules as source code to a file.

        :param file_path: The path to the file to write the source code to.
        :param postfix: The postfix to add to the file name.
        """
        self.generated_python_file_name = self._default_generated_python_file_name + postfix
        func_def = f"def classify(case: {self.case_type.__name__}) -> {self.conclusion_type_hint}:\n"
        file_name = file_path + f"/{self.generated_python_file_name}.py"
        defs_file_name = file_path + f"/{self.generated_python_defs_file_name}.py"
        imports, defs_imports = self._get_imports()
        # clear the files first
        with open(defs_file_name, "w") as f:
            f.write(defs_imports + "\n\n")
        with open(file_name, "w") as f:
            imports += f"from .{self.generated_python_defs_file_name} import *\n"
            imports += f"from ripple_down_rules.rdr import {self.__class__.__name__}\n"
            f.write(imports + "\n\n")
            f.write(f"conclusion_type = ({', '.join([ct.__name__ for ct in self.conclusion_type])},)\n")
            f.write(f"type_ = {self.__class__.__name__}\n")
            f.write(f"\n\n{func_def}")
            f.write(f"{' ' * 4}if not isinstance(case, Case):\n"
                    f"{' ' * 4}    case = create_case(case, max_recursion_idx=3)\n""")
            self.write_rules_as_source_code_to_file(self.start_rule, f, " " * 4, defs_file=defs_file_name)

    @property
    @abstractmethod
    def conclusion_type_hint(self) -> str:
        """
        :return: The type hint of the conclusion of the rdr as a string.
        """
        pass

    def _get_imports(self) -> Tuple[str, str]:
        """
        :return: The imports for the generated python file of the RDR as a string.
        """
        defs_imports = ""
        for rule in [self.start_rule] + list(self.start_rule.descendants):
            if not rule.conditions:
                continue
            for scope in [rule.conditions.scope, rule.conclusion.scope]:
                if scope is None:
                    continue
                for k, v in scope.items():
                    if not hasattr(v, "__module__") or not hasattr(v, "__name__"):
                        continue
                    new_imports = f"from {v.__module__} import {v.__name__}\n"
                    if new_imports in defs_imports:
                        continue
                    defs_imports += new_imports
        imports = ""
        if self.case_type.__module__ != "builtins":
            new_import = f"from {self.case_type.__module__} import {self.case_type.__name__}\n"
            if new_import not in defs_imports:
                imports += new_import
        for conclusion_type in self.conclusion_type:
            if conclusion_type.__module__ != "builtins":
                new_import = f"from {conclusion_type.__module__} import {conclusion_type.__name__}\n"
                if new_import not in defs_imports:
                    imports += new_import
        imports += "from ripple_down_rules.datastructures.case import Case, create_case\n"
        return imports, defs_imports

    def get_rdr_classifier_from_python_file(self, package_name: str) -> Callable[[Any], Any]:
        """
        :param package_name: The name of the package that contains the RDR classifier function.
        :return: The module that contains the rdr classifier function.
        """
        # remove from imports if exists first
        name = f"{package_name.strip('./')}.{self.generated_python_file_name}"
        try:
            module = importlib.import_module(name)
            del sys.modules[name]
        except ModuleNotFoundError:
            pass
        return importlib.import_module(name).classify

    @property
    def generated_python_file_name(self) -> str:
        if self._generated_python_file_name is None:
            self._generated_python_file_name = self._default_generated_python_file_name
        return self._generated_python_file_name

    @generated_python_file_name.setter
    def generated_python_file_name(self, value: str):
        """
        Set the generated python file name.
        :param value: The new value for the generated python file name.
        """
        self._generated_python_file_name = value

    @property
    def _default_generated_python_file_name(self) -> str:
        """
        :return: The default generated python file name.
        """
        if isinstance(self.start_rule.corner_case, Case):
            name = self.start_rule.corner_case._name
        else:
            name = self.start_rule.corner_case.__class__.__name__
        return f"{name.lower()}_{self.attribute_name}_{self.acronym.lower()}"

    @property
    def generated_python_defs_file_name(self) -> str:
        return f"{self.generated_python_file_name}_defs"

    @property
    def acronym(self) -> str:
        """
        :return: The acronym of the classifier.
        """
        if self.__class__.__name__ == "GeneralRDR":
            return "GRDR"
        elif self.__class__.__name__ == "MultiClassRDR":
            return "MCRDR"
        else:
            return "SCRDR"

    @property
    def case_type(self) -> Type:
        """
        :return: The type of the case (input) to the RDR classifier.
        """
        if isinstance(self.start_rule.corner_case, Case):
            return self.start_rule.corner_case._obj_type
        else:
            return type(self.start_rule.corner_case)

    @property
    def conclusion_type(self) -> Tuple[Type]:
        """
        :return: The type of the conclusion of the RDR classifier.
        """
        all_types = []
        for rule in [self.start_rule] + list(self.start_rule.descendants):
            all_types.extend(list(rule.conclusion.conclusion_type))
        return tuple(set(all_types))

    @property
    def attribute_name(self) -> str:
        """
        :return: The name of the attribute that the classifier is classifying.
        """
        return self.start_rule.conclusion_name


class SingleClassRDR(RDRWithCodeWriter):

    def __init__(self, start_rule: Optional[SingleClassRule] = None, default_conclusion: Optional[Any] = None):
        """
        :param start_rule: The starting rule for the classifier.
        :param default_conclusion: The default conclusion for the classifier if no rules fire.
        """
        super(SingleClassRDR, self).__init__(start_rule)
        self.default_conclusion: Optional[Any] = default_conclusion

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CaseAttribute, CallableExpression, None]:
        """
        Classify a case, and ask the user for refinements or alternatives if the classification is incorrect by
        comparing the case with the target category if provided.

        :param case_query: The case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        if case_query.default_value is not None and self.default_conclusion != case_query.default_value:
            self.default_conclusion = case_query.default_value

        pred = self.evaluate(case_query.case)
        if pred.conclusion(case_query.case) != case_query.target_value:
            expert.ask_for_conditions(case_query, pred)
            pred.fit_rule(case_query.case, case_query.target, conditions=case_query.conditions)

        return self.classify(case_query.case)

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if not self.start_rule:
            expert.ask_for_conditions(case_query)
            self.start_rule = SingleClassRule(case_query.conditions, case_query.target, corner_case=case_query.case,
                                              conclusion_name=case_query.attribute_name)

    def classify(self, case: Case, modify_original_case: bool = False) -> Optional[Any]:
        """
        Classify a case by recursively evaluating the rules until a rule fires or the last rule is reached.

        :param case: The case to classify.
        :param modify_original_case: Whether to modify the original case attributes with the conclusion or not.
        """
        pred = self.evaluate(case)
        return pred.conclusion(case) if pred is not None and pred.fired else self.default_conclusion

    def evaluate(self, case: Case) -> SingleClassRule:
        """
        Evaluate the starting rule on a case.
        """
        matched_rule = self.start_rule(case) if self.start_rule is not None else None
        return matched_rule if matched_rule is not None else self.start_rule

    def write_to_python_file(self, file_path: str, postfix: str = ""):
        super().write_to_python_file(file_path, postfix)
        if self.default_conclusion is not None:
            with open(file_path + f"/{self.generated_python_file_name}.py", "a") as f:
                f.write(f"{' ' * 4}else:\n{' ' * 8}return {self.default_conclusion}\n")

    def write_rules_as_source_code_to_file(self, rule: SingleClassRule, file: TextIOWrapper, parent_indent: str = "",
                                           defs_file: Optional[str] = None):
        """
        Write the rules as source code to a file.
        """
        if rule.conditions:
            if_clause = rule.write_condition_as_source_code(parent_indent, defs_file)
            file.write(if_clause)
            if rule.refinement:
                self.write_rules_as_source_code_to_file(rule.refinement, file, parent_indent + "    ",
                                                        defs_file=defs_file)

            conclusion_call = rule.write_conclusion_as_source_code(parent_indent, defs_file)
            file.write(conclusion_call)

            if rule.alternative:
                self.write_rules_as_source_code_to_file(rule.alternative, file, parent_indent, defs_file=defs_file)

    @property
    def conclusion_type_hint(self) -> str:
        return self.conclusion_type[0].__name__

    @property
    def conclusion_type(self) -> Tuple[Type]:
        if self.default_conclusion is not None:
            return (type(self.default_conclusion),)
        return super().conclusion_type

    def _to_json(self) -> Dict[str, Any]:
        return {"start_rule": self.start_rule.to_json()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create an instance of the class from a json
        """
        start_rule = SingleClassRule.from_json(data["start_rule"])
        return cls(start_rule)


class MultiClassRDR(RDRWithCodeWriter):
    """
    A multi class ripple down rules classifier, which can draw multiple conclusions for a case.
    This is done by going through all rules and checking if they fire or not, and adding stopping rules if needed,
    when wrong conclusions are made to stop these rules from firing again for similar cases.
    """
    evaluated_rules: Optional[List[Rule]] = None
    """
    The evaluated rules in the classifier for one case.
    """
    conclusions: Optional[List[CaseAttribute]] = None
    """
    The conclusions that the case belongs to.
    """
    stop_rule_conditions: Optional[CallableExpression] = None
    """
    The conditions of the stopping rule if needed.
    """

    def __init__(self, start_rule: Optional[MultiClassTopRule] = None,
                 mode: MCRDRMode = MCRDRMode.StopOnly):
        """
        :param start_rule: The starting rules for the classifier.
        :param mode: The mode of the classifier, either StopOnly or StopPlusRule, or StopPlusRuleCombined.
        """
        super(MultiClassRDR, self).__init__(start_rule)
        self.mode: MCRDRMode = mode

    def classify(self, case: Union[Case, SQLTable]) -> Set[Any]:
        evaluated_rule = self.start_rule
        self.conclusions = []
        while evaluated_rule:
            next_rule = evaluated_rule(case)
            if evaluated_rule.fired:
                self.add_conclusion(evaluated_rule, case)
            evaluated_rule = next_rule
        return make_set(self.conclusions)

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None
                 , **kwargs) -> Set[Union[CaseAttribute, CallableExpression, None]]:
        """
        Classify a case, and ask the user for stopping rules or classifying rules if the classification is incorrect
         or missing by comparing the case with the target category if provided.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions or for extra conclusions.
        :return: The conclusions that the case belongs to.
        """
        self.conclusions = []
        self.stop_rule_conditions = None
        evaluated_rule = self.start_rule
        target = make_set(case_query.target_value)
        while evaluated_rule:
            next_rule = evaluated_rule(case_query.case)
            rule_conclusion = evaluated_rule.conclusion(case_query.case)

            if evaluated_rule.fired:
                if not make_set(rule_conclusion).issubset(target):
                    # Rule fired and conclusion is different from target
                    self.stop_wrong_conclusion_else_add_it(case_query, expert, evaluated_rule)
                else:
                    # Rule fired and target is correct or there is no target to compare
                    self.add_conclusion(evaluated_rule, case_query.case)

            if not next_rule:
                if not make_set(target).issubset(make_set(self.conclusions)):
                    # Nothing fired and there is a target that should have been in the conclusions
                    self.add_rule_for_case(case_query, expert)
                    # Have to check all rules again to make sure only this new rule fires
                    next_rule = self.start_rule
            evaluated_rule = next_rule
        return self.conclusions

    def write_rules_as_source_code_to_file(self, rule: Union[MultiClassTopRule, MultiClassStopRule],
                                           file, parent_indent: str = "", defs_file: Optional[str] = None):
        if rule == self.start_rule:
            file.write(f"{parent_indent}conclusions = set()\n")
        if rule.conditions:
            if_clause = rule.write_condition_as_source_code(parent_indent, defs_file)
            file.write(if_clause)
            conclusion_indent = parent_indent
            if hasattr(rule, "refinement") and rule.refinement:
                self.write_rules_as_source_code_to_file(rule.refinement, file, parent_indent + "    ",
                                                        defs_file=defs_file)
                conclusion_indent = parent_indent + " " * 4
                file.write(f"{conclusion_indent}else:\n")

            conclusion_call = rule.write_conclusion_as_source_code(conclusion_indent, defs_file)
            file.write(conclusion_call)

            if rule.alternative:
                self.write_rules_as_source_code_to_file(rule.alternative, file, parent_indent, defs_file=defs_file)

    @property
    def conclusion_type_hint(self) -> str:
        conclusion_types = [ct.__name__ for ct in self.conclusion_type if ct not in [list, set]]
        if len(conclusion_types) == 1:
            return f"Set[{conclusion_types[0]}]"
        else:
            return f"Set[Union[{', '.join(conclusion_types)}]]"

    def _get_imports(self) -> Tuple[str, str]:
        imports, defs_imports = super()._get_imports()
        conclusion_types = [ct for ct in self.conclusion_type if ct not in [list, set]]
        if len(conclusion_types) == 1:
            imports += f"from typing_extensions import Set\n"
        else:
            imports += "from typing_extensions import Set, Union\n"
        imports += "from ripple_down_rules.utils import make_set\n"
        return imports, defs_imports

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if not self.start_rule:
            conditions = expert.ask_for_conditions(case_query)
            self.start_rule = MultiClassTopRule(conditions, case_query.target, corner_case=case_query.case,
                                                conclusion_name=case_query.attribute_name)

    @property
    def last_top_rule(self) -> Optional[MultiClassTopRule]:
        """
        Get the last top rule in the tree.
        """
        if not self.start_rule.furthest_alternative:
            return self.start_rule
        else:
            return self.start_rule.furthest_alternative[-1]

    def stop_wrong_conclusion_else_add_it(self, case_query: CaseQuery, expert: Expert,
                                          evaluated_rule: MultiClassTopRule):
        """
        Stop a wrong conclusion by adding a stopping rule.
        """
        rule_conclusion = evaluated_rule.conclusion(case_query.case)
        if is_conflicting(rule_conclusion, case_query.target_value):
            self.stop_conclusion(case_query, expert, evaluated_rule)

    def stop_conclusion(self, case_query: CaseQuery,
                        expert: Expert, evaluated_rule: MultiClassTopRule):
        """
        Stop a conclusion by adding a stopping rule.

        :param case_query: The case query to stop the conclusion for.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param evaluated_rule: The evaluated rule to ask the expert about.
        """
        conditions = expert.ask_for_conditions(case_query, evaluated_rule)
        evaluated_rule.fit_rule(case_query.case, case_query.target, conditions=conditions)
        if self.mode == MCRDRMode.StopPlusRule:
            self.stop_rule_conditions = conditions
        if self.mode == MCRDRMode.StopPlusRuleCombined:
            new_top_rule_conditions = conditions.combine_with(evaluated_rule.conditions)
            self.add_top_rule(new_top_rule_conditions, case_query.target, case_query.case)

    def add_rule_for_case(self, case_query: CaseQuery, expert: Expert):
        """
        Add a rule for a case that has not been classified with any conclusion.

        :param case_query: The case query to add the rule for.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if self.stop_rule_conditions and self.mode == MCRDRMode.StopPlusRule:
            conditions = self.stop_rule_conditions
            self.stop_rule_conditions = None
        else:
            conditions = expert.ask_for_conditions(case_query)
        self.add_top_rule(conditions, case_query.target, case_query.case)

    def add_conclusion(self, evaluated_rule: Rule, case: Case) -> None:
        """
        Add the conclusion of the evaluated rule to the list of conclusions.

        :param evaluated_rule: The evaluated rule to add the conclusion of.
        :param case: The case to add the conclusion for.
        """
        conclusion_types = [type(c) for c in self.conclusions]
        rule_conclusion = evaluated_rule.conclusion(case)
        if type(rule_conclusion) not in conclusion_types:
            self.conclusions.extend(make_list(rule_conclusion))
        else:
            same_type_conclusions = [c for c in self.conclusions if type(c) == type(rule_conclusion)]
            combined_conclusion = rule_conclusion if isinstance(rule_conclusion, set) \
                else {rule_conclusion}
            combined_conclusion = copy(combined_conclusion)
            for c in same_type_conclusions:
                combined_conclusion.update(c if isinstance(c, set) else make_set(c))
                self.conclusions.remove(c)
            self.conclusions.extend(make_list(combined_conclusion))

    def add_top_rule(self, conditions: CallableExpression, conclusion: Any, corner_case: Union[Case, SQLTable]):
        """
        Add a top rule to the classifier, which is a rule that is always checked and is part of the start_rules list.

        :param conditions: The conditions of the rule.
        :param conclusion: The conclusion of the rule.
        :param corner_case: The corner case of the rule.
        """
        self.start_rule.alternative = MultiClassTopRule(conditions, conclusion, corner_case=corner_case)

    def _to_json(self) -> Dict[str, Any]:
        return {"start_rule": self.start_rule.to_json()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create an instance of the class from a json
        """
        start_rule = SingleClassRule.from_json(data["start_rule"])
        return cls(start_rule)


class GeneralRDR(RippleDownRules):
    """
    A general ripple down rules classifier, which can draw multiple conclusions for a case, but each conclusion is part
    of a set of mutually exclusive conclusions. Whenever a conclusion is made, the classification restarts from the
    starting rule, and all the rules that belong to the class of the made conclusion are not checked again. This
    continues until no more rules can be fired. In addition, previous conclusions can be used as conditions or input to
    the next classification/cycle.
    Another possible mode is to have rules that are considered final, when fired, inference will not be restarted,
     and only a refinement can be made to the final rule, those can also be used in another SCRDR of their own that
     gets called when the final rule fires.
    """

    def __init__(self, category_rdr_map: Optional[Dict[str, Union[SingleClassRDR, MultiClassRDR]]] = None):
        """
        :param category_rdr_map: A map of case attribute names to ripple down rules classifiers,
        where each category is a parent category that has a set of mutually exclusive (in case of SCRDR) child
        categories, e.g. {'species': SCRDR, 'habitats': MCRDR}, where 'species' and 'habitats' are attribute names
        for a case of type Animal, while SCRDR and MCRDR are SingleClass and MultiClass ripple down rules classifiers.
        Species can have values like Mammal, Bird, Fish, etc. which are mutually exclusive, while Habitat can have
        values like Land, Water, Air, etc., which are not mutually exclusive due to some animals living more than one
        habitat.
        """
        self.start_rules_dict: Dict[str, Union[SingleClassRDR, MultiClassRDR]] \
            = category_rdr_map if category_rdr_map else {}
        super(GeneralRDR, self).__init__()
        self.all_figs: List[plt.Figure] = [sr.fig for sr in self.start_rules_dict.values()]

    def add_rdr(self, rdr: Union[SingleClassRDR, MultiClassRDR], attribute_name: Optional[str] = None):
        """
        Add a ripple down rules classifier to the map of classifiers.

        :param rdr: The ripple down rules classifier to add.
        :param attribute_name: The name of the attribute that the classifier is classifying.
        """
        attribute_name = attribute_name if attribute_name else rdr.attribute_name
        self.start_rules_dict[attribute_name] = rdr

    @property
    def start_rule(self) -> Optional[Union[SingleClassRule, MultiClassTopRule]]:
        return self.start_rules[0] if self.start_rules_dict else None

    @start_rule.setter
    def start_rule(self, value: Union[SingleClassRDR, MultiClassRDR]):
        if value:
            self.start_rules_dict[value.attribute_name] = value

    @property
    def start_rules(self) -> List[Union[SingleClassRule, MultiClassTopRule]]:
        return [rdr.start_rule for rdr in self.start_rules_dict.values()]

    def classify(self, case: Any, modify_case: bool = False) -> Optional[Dict[str, Any]]:
        """
        Classify a case by going through all RDRs and adding the categories that are classified, and then restarting
        the classification until no more categories can be added.

        :param case: The case to classify.
        :param modify_case: Whether to modify the original case or create a copy and modify it.
        :return: The categories that the case belongs to.
        """
        return self._classify(self.start_rules_dict, case, modify_original_case=modify_case)

    @staticmethod
    def _classify(classifiers_dict: Dict[str, Union[ModuleType, RippleDownRules]],
                  case: Any, modify_original_case: bool = False) -> Dict[str, Any]:
        """
        Classify a case by going through all classifiers and adding the categories that are classified,
         and then restarting the classification until no more categories can be added.

        :param classifiers_dict: A dictionary mapping conclusion types to the classifiers that produce them.
        :param case: The case to classify.
        :param modify_original_case: Whether to modify the original case or create a copy and modify it.
        :return: The categories that the case belongs to.
        """
        conclusions = {}
        case = case if isinstance(case, (Case, SQLTable)) else create_case(case)
        case_cp = copy_case(case) if not modify_original_case else case
        while True:
            new_conclusions = {}
            for attribute_name, rdr in classifiers_dict.items():
                pred_atts = rdr.classify(case_cp)
                if pred_atts is None:
                    continue
                if rdr.type_ is SingleClassRDR:
                    if attribute_name not in conclusions or \
                            (attribute_name in conclusions and conclusions[attribute_name] != pred_atts):
                        conclusions[attribute_name] = pred_atts
                        new_conclusions[attribute_name] = pred_atts
                else:
                    pred_atts = make_set(pred_atts)
                    if attribute_name in conclusions:
                        pred_atts = {p for p in pred_atts if p not in conclusions[attribute_name]}
                    if len(pred_atts) > 0:
                        new_conclusions[attribute_name] = pred_atts
                        if attribute_name not in conclusions:
                            conclusions[attribute_name] = set()
                        conclusions[attribute_name].update(pred_atts)
                if attribute_name in new_conclusions:
                    mutually_exclusive = True if rdr.type_ is SingleClassRDR else False
                    case_query = CaseQuery(case_cp, attribute_name, rdr.conclusion_type, mutually_exclusive)
                    update_case(case_query, new_conclusions)
            if len(new_conclusions) == 0:
                break
        return conclusions

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Dict[str, Any]:
        """
        Fit the GRDR on a case, if the target is a new type of category, a new RDR is created for it,
        else the existing RDR of that type will be fitted on the case, and then classification is done and all
        concluded categories are returned. If the category is mutually exclusive, an SCRDR is created, else an MCRDR.
        In case of SCRDR, multiple conclusions of the same type replace each other, in case of MCRDR, they are added if
        they are accepted by the expert, and the attribute of that category is represented in the case as a set of
        values.

        :param case_query: The query containing the case to classify and the target category to compare the case
        with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The categories that the case belongs to.
        """

        case_query_cp = copy(case_query)
        self.classify(case_query_cp.case, modify_case=True)
        case_query_cp.update_target_value()

        self.start_rules_dict[case_query_cp.attribute_name].fit_case(case_query_cp, expert, **kwargs)

        return self.classify(case_query.case)

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if case_query.attribute_name not in self.start_rules_dict:
            new_rdr = self.initialize_new_rdr_for_attribute(case_query)
            self.add_rdr(new_rdr, case_query.attribute_name)

    @staticmethod
    def initialize_new_rdr_for_attribute(case_query: CaseQuery):
        """
        Initialize the appropriate RDR type for the target.
        """
        return SingleClassRDR(default_conclusion=case_query.default_value) if case_query.mutually_exclusive \
            else MultiClassRDR()

    def _to_json(self) -> Dict[str, Any]:
        return {"start_rules": {t: rdr.to_json() for t, rdr in self.start_rules_dict.items()}}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> GeneralRDR:
        """
        Create an instance of the class from a json
        """
        start_rules_dict = {}
        for k, v in data["start_rules"].items():
            start_rules_dict[k] = get_type_from_string(v['_type']).from_json(v)
        return cls(start_rules_dict)

    def write_to_python_file(self, file_path: str, postfix: str = "") -> None:
        """
        Write the tree of rules as source code to a file.

        :param file_path: The path to the file to write the source code to.
        :param postfix: The postfix to add to the file name.
        """
        self.generated_python_file_name = self._default_generated_python_file_name + postfix
        for rdr in self.start_rules_dict.values():
            rdr.write_to_python_file(file_path, postfix=f"_of_grdr{postfix}")
        func_def = f"def classify(case: {self.case_type.__name__}) -> {self.conclusion_type_hint}:\n"
        with open(file_path + f"/{self.generated_python_file_name}.py", "w") as f:
            f.write(self._get_imports(file_path) + "\n\n")
            f.write("classifiers_dict = dict()\n")
            for rdr_key, rdr in self.start_rules_dict.items():
                f.write(f"classifiers_dict['{rdr_key}'] = {self.rdr_key_to_function_name(rdr_key)}\n")
            f.write("\n\n")
            f.write(func_def)
            f.write(f"{' ' * 4}if not isinstance(case, Case):\n"
                    f"{' ' * 4}    case = create_case(case, max_recursion_idx=3)\n""")
            f.write(f"{' ' * 4}return GeneralRDR._classify(classifiers_dict, case)\n")

    @property
    def case_type(self) -> Type:
        """
        :return: The type of the case (input) to the RDR classifier.
        """
        if isinstance(self.start_rule.corner_case, Case):
            return self.start_rule.corner_case._obj_type
        else:
            return type(self.start_rule.corner_case)

    def get_rdr_classifier_from_python_file(self, file_path: str) -> Callable[[Any], Any]:
        """
        :param file_path: The path to the file that contains the RDR classifier function.
        :param postfix: The postfix to add to the file name.
        :return: The module that contains the rdr classifier function.
        """
        return importlib.import_module(f"{file_path.strip('./')}.{self.generated_python_file_name}").classify

    @property
    def generated_python_file_name(self) -> str:
        if self._generated_python_file_name is None:
            self._generated_python_file_name = self._default_generated_python_file_name
        return self._generated_python_file_name

    @generated_python_file_name.setter
    def generated_python_file_name(self, value: str):
        self._generated_python_file_name = value

    @property
    def _default_generated_python_file_name(self) -> str:
        """
        :return: The default generated python file name.
        """
        if isinstance(self.start_rule.corner_case, Case):
            name = self.start_rule.corner_case._name
        else:
            name = self.start_rule.corner_case.__class__.__name__
        return f"{name}_rdr".lower()

    @property
    def conclusion_type_hint(self) -> str:
        return f"Dict[str, Any]"

    def _get_imports(self, file_path: str) -> str:
        """
        Get the imports needed for the generated python file.

        :param file_path: The path to the file that contains the RDR classifier function.
        :return: The imports needed for the generated python file.
        """
        imports = ""
        # add type hints
        imports += f"from typing_extensions import Dict, Any\n"
        # import rdr type
        imports += f"from ripple_down_rules.rdr import GeneralRDR\n"
        # add case type
        imports += f"from ripple_down_rules.datastructures.case import Case, create_case\n"
        imports += f"from {self.case_type.__module__} import {self.case_type.__name__}\n"
        # add rdr python generated functions.
        for rdr_key, rdr in self.start_rules_dict.items():
            imports += (f"from {file_path.strip('./')}"
                        f" import {rdr.generated_python_file_name} as {self.rdr_key_to_function_name(rdr_key)}\n")
        return imports

    @staticmethod
    def rdr_key_to_function_name(rdr_key: str) -> str:
        """
        Convert the RDR key to a function name.

        :param rdr_key: The RDR key to convert.
        :return: The function name.
        """
        return rdr_key.replace(".", "_").lower() + "_classifier"



import importlib
import os
import time
from unittest import TestCase, skip

from typing_extensions import List

from ripple_down_rules.datasets import Habitat, Species
from ripple_down_rules.datasets import load_zoo_dataset
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import MCRDRMode
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR, MultiClassRDR, GeneralRDR
from ripple_down_rules.utils import render_tree, get_all_subclasses, make_set, flatten_list
from test_helpers.helpers import get_fit_scrdr, get_fit_mcrdr, get_fit_grdr, get_habitat


class TestRDR(TestCase):
    all_cases: List[Case]
    targets: List[str]
    case_queries: List[CaseQuery]
    test_results_dir: str = "./test_results"
    expert_answers_dir: str = "./test_expert_answers"
    generated_rdrs_dir: str = "./test_generated_rdrs"
    cache_file: str = f"{test_results_dir}/zoo_dataset.pkl"

    @classmethod
    def setUpClass(cls):
        # fetch dataset
        cls.all_cases, cls.targets = load_zoo_dataset(cache_file=cls.cache_file)
        cls.case_queries = [CaseQuery(case, "species", Species, True, _target=target)
                            for case, target in zip(cls.all_cases, cls.targets)]
        for test_dir in [cls.test_results_dir, cls.expert_answers_dir, cls.generated_rdrs_dir]:
            if not os.path.exists(test_dir):
                os.makedirs(test_dir)

    def test_classify_scrdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/scrdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        scrdr = SingleClassRDR()
        cat = scrdr.fit_case(self.case_queries[0], expert=expert)
        self.assertEqual(cat, self.targets[0])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_scrdr(self):
        scrdr, _ = get_fit_scrdr(self.all_cases, self.targets, draw_tree=False,
                                 expert_answers_dir=self.expert_answers_dir,
                                 expert_answers_file="scrdr_expert_answers_fit",
                                 load_answers=True)
        render_tree(scrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/scrdr")

    def test_fit_scrdr_with_no_targets(self):
        # Test with no targets
        scrdr, case_queries = get_fit_scrdr(self.all_cases[:20], [], draw_tree=False,
                                            expert_answers_dir=self.expert_answers_dir,
                                            expert_answers_file="scrdr_expert_answers_fit_no_targets",
                                            load_answers=True,
                                            save_answers=False)
        render_tree(scrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/scrdr_no_targets")

    def test_write_scrdr_no_targets_to_python_file(self):
        # Test with no targets
        scrdr, case_queries = get_fit_scrdr(self.all_cases[:20], [], draw_tree=False,
                                            expert_answers_dir=self.expert_answers_dir,
                                            expert_answers_file="scrdr_expert_answers_fit_no_targets",
                                            load_answers=True, save_answers=False)
        scrdr.write_to_python_file(self.generated_rdrs_dir, postfix="_no_targets")
        classify_species_scrdr = scrdr.get_rdr_classifier_from_python_file(self.generated_rdrs_dir)
        for case_query, target in zip(case_queries, self.targets):
            cat = classify_species_scrdr(case_query.case)
            self.assertEqual(cat, target)

    def test_fit_mcrdr_with_no_targets(self):
        # Test with no targets
        mcrdr = get_fit_mcrdr(self.all_cases[:20], [], draw_tree=False,
                              expert_answers_dir=self.expert_answers_dir,
                              expert_answers_file="mcrdr_expert_answers_fit_no_targets",
                              load_answers=True,
                              save_answers=False)
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_no_targets")
        for case, target in zip(self.all_cases[:20], self.targets[:20]):
            cat = mcrdr.classify(case)
            self.assertEqual(make_set(cat), make_set(target))

    def test_write_mcrdr_no_targets_to_python_file(self):
        # Test with no targets
        mcrdr = get_fit_mcrdr(self.all_cases[:20], [], draw_tree=False,
                              expert_answers_dir=self.expert_answers_dir,
                              expert_answers_file="mcrdr_expert_answers_fit_no_targets",
                              load_answers=True, save_answers=False)
        mcrdr.write_to_python_file(self.generated_rdrs_dir, postfix="_no_targets")
        classify_species_mcrdr = mcrdr.get_rdr_classifier_from_python_file(self.generated_rdrs_dir)
        for case, target in zip(self.all_cases[:20], self.targets[:20]):
            cat = classify_species_mcrdr(case)
            self.assertEqual(make_set(cat), make_set(target))

    # @skip("Test is not implemented yet")
    def test_fit_grdr_with_no_targets(self):
        draw_tree = False
        # Test with no targets
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets, draw_tree=draw_tree,
                                         expert_answers_dir=self.expert_answers_dir,
                                         expert_answers_file="grdr_expert_answers_fit_no_targets",
                                         load_answers=True, save_answers=False, append=False, no_targets=True)
        if draw_tree:
            for conclusion_name, rdr in grdr.start_rules_dict.items():
                render_tree(rdr.start_rule, use_dot_exporter=True,
                            filename=self.test_results_dir + f"/grdr_no_targets_{conclusion_name}")
        for case, case_targets in zip(self.all_cases[:20], all_targets):
            cat = grdr.classify(case)
            for cat_name, cat_val in cat.items():
                if cat_name in case_targets:
                    self.assertEqual(make_set(cat_val), make_set(case_targets[cat_name]))

    def test_write_grdr_no_targets_to_python_file(self):
        # Test with no targets
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets, draw_tree=False,
                                         expert_answers_dir=self.expert_answers_dir,
                                         expert_answers_file="grdr_expert_answers_fit_no_targets",
                                         load_answers=True, save_answers=False, append=False, no_targets=True)
        grdr.write_to_python_file(self.generated_rdrs_dir, postfix="_no_targets")
        classify_species_grdr = grdr.get_rdr_classifier_from_python_file(self.generated_rdrs_dir)
        for case, case_targets in zip(self.all_cases[:20], all_targets):
            cat = classify_species_grdr(case)
            for cat_name, cat_val in cat.items():
                if cat_name in case_targets:
                    self.assertEqual(make_set(cat_val), make_set(case_targets[cat_name]))

    def test_fit_multi_line_scrdr(self):
        n = 20
        scrdr, _ = get_fit_scrdr(self.all_cases[:n], self.targets[:n], draw_tree=False,
                                 expert_answers_dir=self.expert_answers_dir,
                                 expert_answers_file="scrdr_multi_line_expert_answers_fit",
                                 load_answers=True,
                                 save_answers=False)
        render_tree(scrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/scrdr_multi_line")

    def test_write_multi_line_scrdr_to_python_file(self):
        n = 20
        scrdr, _ = get_fit_scrdr(self.all_cases[:n], self.targets[:n], draw_tree=False,
                                 expert_answers_dir=self.expert_answers_dir,
                                 expert_answers_file="scrdr_multi_line_expert_answers_fit",
                                 load_answers=True)
        scrdr.write_to_python_file(self.generated_rdrs_dir, postfix="_multi_line")
        classify_species_scrdr = scrdr.get_rdr_classifier_from_python_file(self.generated_rdrs_dir)
        for case, target in zip(self.all_cases[:n], self.targets[:n]):
            cat = classify_species_scrdr(case)
            self.assertEqual(cat, target)

    def test_write_scrdr_to_python_file(self):
        scrdr, _ = get_fit_scrdr(self.all_cases, self.targets)
        scrdr.write_to_python_file(self.generated_rdrs_dir)
        classify_species_scrdr = scrdr.get_rdr_classifier_from_python_file(self.generated_rdrs_dir)
        for case, target in zip(self.all_cases, self.targets):
            cat = classify_species_scrdr(case)
            self.assertEqual(cat, target)

    def test_write_mcrdr_to_python_file(self):
        mcrdr = get_fit_mcrdr(self.all_cases, self.targets)
        mcrdr.write_to_python_file(self.generated_rdrs_dir)
        classify_species_mcrdr = mcrdr.get_rdr_classifier_from_python_file(self.generated_rdrs_dir)
        for case, target in zip(self.all_cases, self.targets):
            cat = classify_species_mcrdr(case)
            self.assertEqual(make_set(cat), make_set(target))

    def test_write_mcrdr_multi_line_to_python_file(self):
        n = 20
        mcrdr = get_fit_mcrdr(self.all_cases[:n], self.targets[:n], draw_tree=False,
                              expert_answers_dir=self.expert_answers_dir,
                              expert_answers_file="mcrdr_multi_line_expert_answers_fit",
                              load_answers=True,
                              save_answers=False)
        mcrdr.write_to_python_file(self.generated_rdrs_dir, postfix="_multi_line")
        classify_species_mcrdr = mcrdr.get_rdr_classifier_from_python_file(self.generated_rdrs_dir)
        for case, target in zip(self.all_cases[:n], self.targets[:n]):
            cat = classify_species_mcrdr(case)
            self.assertEqual(make_set(cat), make_set(target))

    def test_write_grdr_to_python_file(self):
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets)
        grdr.write_to_python_file(self.generated_rdrs_dir)
        classify_animal_grdr = grdr.get_rdr_classifier_from_python_file(self.generated_rdrs_dir)
        for case, case_targets in zip(self.all_cases[:len(all_targets)], all_targets):
            cat = classify_animal_grdr(case)
            for cat_name, cat_val in cat.items():
                self.assertEqual(make_set(cat_val), make_set(case_targets[cat_name]))

    def test_classify_mcrdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/mcrdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        mcrdr = MultiClassRDR()
        cats = mcrdr.fit_case(self.case_queries[0],
                              expert=expert)

        self.assertEqual(cats[0], self.targets[0])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_mcrdr_stop_only(self):
        use_loaded_answers = True
        draw_tree = False
        save_answers = False
        filename = self.expert_answers_dir + "/mcrdr_expert_answers_stop_only_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR()
        case_queries = self.case_queries
        mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_stop_only")
        for case_query in case_queries:
            cat = mcrdr.classify(case_query.case)
            self.assertEqual(make_set(cat), make_set(case_query.target_value))
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_mcrdr_stop_plus_rule(self):
        use_loaded_answers = True
        draw_tree = False
        save_answers = False
        append = False
        filename = self.expert_answers_dir + "/mcrdr_stop_plus_rule_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers, append=append)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR(mode=MCRDRMode.StopPlusRule)
        case_queries = self.case_queries
        mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_stop_plus_rule")
        for case_query in case_queries:
            cat = mcrdr.classify(case_query.case)
            self.assertEqual(make_set(cat), make_set(case_query.target_value))
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_mcrdr_stop_plus_rule_combined(self):
        use_loaded_answers = True
        save_answers = False
        draw_tree = False
        append = False
        filename = self.expert_answers_dir + "/mcrdr_stop_plus_rule_combined_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers, append=append)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR(mode=MCRDRMode.StopPlusRuleCombined)
        case_queries = self.case_queries
        mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        render_tree(mcrdr.start_rule, use_dot_exporter=True,
                    filename=self.test_results_dir + f"/mcrdr_stop_plus_rule_combined")
        for case_query in case_queries:
            cat = mcrdr.classify(case_query.case)
            self.assertEqual(make_set(cat), make_set(case_query.target_value))
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_classify_grdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/grdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        grdr = GeneralRDR()

        targets = [self.targets[0], Habitat.land]
        attribute_names = [t.__class__.__name__.lower() for t in targets]
        targets = dict(zip(attribute_names, targets))
        case_queries = [CaseQuery(self.all_cases[0], a, (type(t),), True if a == 'species' else False,
                                  _target=t) for a, t in targets.items()]
        grdr.fit(case_queries, expert=expert)
        cats = grdr.classify(self.all_cases[0])
        for cat_name, value in cats.items():
            self.assertEqual(make_set(value), make_set(targets[cat_name]))

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_grdr(self):
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets, draw_tree=False,
                                         load_answers=True, save_answers=False)
        for conclusion_name, rdr in grdr.start_rules_dict.items():
            render_tree(rdr.start_rule, use_dot_exporter=True,
                        filename=self.test_results_dir + f"/grdr_{conclusion_name}")



import os
from dataclasses import dataclass
from enum import Enum
from unittest import skip

from typing_extensions import List

from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR, GeneralRDR


class Element(str, Enum):
    C = "c"
    H = "h"
    O = "o"
    N = "n"
    F = "f"
    B = "b"
    I = "i"

    def __repr__(self):
        return self.name


@dataclass
class Atom:
    element: Element
    type: int
    charge: float

    def __hash__(self):
        return id(self)


@dataclass
class Bond:
    atom1: Atom
    atom2: Atom
    type: int

    def __hash__(self):
        return id(self)


@dataclass
class Molecule:
    ind1: int
    inda: int
    logp: float
    lumo: float
    mutagenic: bool

    atoms: List[Atom]
    bonds: List[Bond]

    def __hash__(self):
        return id(self)


def make_molecule_1() -> Molecule:
    atoms = [Atom(element=Element.C, type=22, charge=-0.117),
             Atom(element=Element.H, type=3, charge=0.142),
             Atom(element=Element.C, type=27, charge=-0.087),
             Atom(element=Element.C, type=27, charge=0.013), Atom(element=Element.C, type=22, charge=-0.117),
             Atom(element=Element.C, type=22, charge=-0.117), Atom(element=Element.H, type=3, charge=0.143),
             Atom(element=Element.H, type=3, charge=0.143), Atom(element=Element.C, type=22, charge=-0.117),
             Atom(element=Element.C, type=22, charge=-0.117), Atom(element=Element.C, type=22, charge=-0.117),
             Atom(element=Element.C, type=22, charge=-0.117), Atom(element=Element.C, type=22, charge=-0.117),
             Atom(element=Element.H, type=3, charge=0.142), Atom(element=Element.H, type=3, charge=0.143),
             Atom(element=Element.H, type=3, charge=0.142), Atom(element=Element.N, type=38, charge=0.812),
             Atom(element=Element.O, type=40, charge=-0.388), Atom(element=Element.O, type=40, charge=-0.388),
             Atom(element=Element.C, type=22, charge=-0.117), Atom(element=Element.C, type=195, charge=-0.087),
             Atom(element=Element.C, type=195, charge=0.013), Atom(element=Element.C, type=22, charge=-0.117),
             Atom(element=Element.H, type=3, charge=0.142), Atom(element=Element.H, type=3, charge=0.143),
             Atom(element=Element.H, type=3, charge=0.142)]
    bonds = [Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[6], type=1),
             Bond(atom1=atoms[16], atom2=atoms[17], type=2),
             Bond(atom1=atoms[0], atom2=atoms[6], type=1),
             Bond(atom1=atoms[21], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[20], type=7),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[3], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[6], type=1),
             Bond(atom1=atoms[16], atom2=atoms[0], type=1),
             Bond(atom1=atoms[16], atom2=atoms[17], type=2),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[20], atom2=atoms[21], type=7),
             Bond(atom1=atoms[20], atom2=atoms[2], type=7),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[0], atom2=atoms[21], type=7),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[2], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[3], type=7),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[6], type=1),
             Bond(atom1=atoms[2], atom2=atoms[3], type=7)]

    molecule = Molecule(ind1=1, inda=0, logp=4.23, lumo=-1.246, mutagenic=True, atoms=atoms, bonds=bonds)
    return molecule


def make_molecule_2() -> Molecule:
    atoms = [Atom(element=Element.C, type=22, charge=-0.128), Atom(element=Element.H, type=3, charge=0.132),
             Atom(element=Element.C, type=29, charge=0.002), Atom(element=Element.C, type=22, charge=-0.128),
             Atom(element=Element.C, type=22, charge=-0.128), Atom(element=Element.C, type=22, charge=-0.128),
             Atom(element=Element.C, type=22, charge=0.202), Atom(element=Element.C, type=22, charge=-0.128),
             Atom(element=Element.H, type=3, charge=0.132), Atom(element=Element.H, type=3, charge=0.132),
             Atom(element=Element.H, type=3, charge=0.132), Atom(element=Element.C, type=22, charge=-0.128),
             Atom(element=Element.H, type=3, charge=0.132), Atom(element=Element.N, type=32, charge=-0.779),
             Atom(element=Element.N, type=38, charge=0.801), Atom(element=Element.O, type=40, charge=-0.398),
             Atom(element=Element.O, type=40, charge=-0.398), Atom(element=Element.H, type=1, charge=0.332),
             Atom(element=Element.H, type=1, charge=0.332), Atom(element=Element.C, type=22, charge=-0.128),
             Atom(element=Element.C, type=29, charge=0.002), Atom(element=Element.C, type=22, charge=-0.128),
             Atom(element=Element.C, type=22, charge=-0.128), Atom(element=Element.H, type=3, charge=0.132),
             Atom(element=Element.H, type=3, charge=0.132), Atom(element=Element.H, type=3, charge=0.132)]

    bonds = [Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[14], atom2=atoms[15], type=2),
             Bond(atom1=atoms[6], atom2=atoms[13], type=1),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[6], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[2], atom2=atoms[2], type=1),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[0], atom2=atoms[2], type=7),
             Bond(atom1=atoms[13], atom2=atoms[17], type=1),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[0], atom2=atoms[0], type=7),
             Bond(atom1=atoms[0], atom2=atoms[14], type=1),
             Bond(atom1=atoms[0], atom2=atoms[2], type=7),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[2], atom2=atoms[0], type=7),
             Bond(atom1=atoms[13], atom2=atoms[17], type=1),
             Bond(atom1=atoms[14], atom2=atoms[15], type=2),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[0], atom2=atoms[1], type=1),
             Bond(atom1=atoms[0], atom2=atoms[6], type=7),
             Bond(atom1=atoms[2], atom2=atoms[0], type=7)]

    molecule = Molecule(ind1=0, inda=0, logp=2.68, lumo=-1.034, mutagenic=False, atoms=atoms, bonds=bonds)
    return molecule


def get_two_molecules_model(draw_tree=False, load_answers=True, save_answers=False,
                            filename="./test_expert_answers/mutagenic_expert_answers"):
    expert = Human(use_loaded_answers=load_answers)
    if load_answers:
        expert.load_answers(filename)

    molecule_1 = make_molecule_1()
    molecule_2 = make_molecule_2()
    case_queries = [CaseQuery(molecule_1, "mutagenic", bool, True, _target=molecule_1.mutagenic),
                    CaseQuery(molecule_2, "mutagenic", bool, True, _target=molecule_2.mutagenic), ]

    rdr = SingleClassRDR()
    rdr.fit(case_queries, expert=expert, animate_tree=draw_tree)

    for case_query in case_queries:
        r = rdr.classify(case_query.case)
        if isinstance(rdr, GeneralRDR):
            assert r[case_query.attribute_name] == case_query.target_value
        else:
            assert r == case_query.target_value

    if save_answers:
        cwd = os.getcwd()
        file = os.path.join(cwd, filename)
        expert.save_answers(file)

    return rdr


def test_two_molecules():
    rdr = get_two_molecules_model(draw_tree=False, load_answers=True, save_answers=False)


def test_serialize_two_molecules_model():
    rdr = get_two_molecules_model()
    filename = "./test_results/mutagenic_rdr.json"
    rdr.save(filename)
    loaded_rdr = type(rdr).load(filename)
    assert rdr.classify(make_molecule_1()) == loaded_rdr.classify(make_molecule_1())
    assert rdr.classify(make_molecule_2()) == loaded_rdr.classify(make_molecule_2())


def test_write_two_molecules_model_to_python():
    rdr = get_two_molecules_model()
    filename = "./test_generated_rdrs"
    rdr.write_to_python_file(filename)
    loaded_rdr = rdr.get_rdr_classifier_from_python_file(filename)
    assert rdr.classify(make_molecule_1()) == loaded_rdr(make_molecule_1())
    assert rdr.classify(make_molecule_2()) == loaded_rdr(make_molecule_2())


if __name__ == '__main__':
    # test_two_molecules()
    test_write_two_molecules_model_to_python()



from __future__ import annotations

import ast
import logging
from _ast import AST
from enum import Enum

from typing_extensions import Type, Optional, Any, List, Union, Tuple, Dict, Set

from .case import create_case, Case
from ..utils import SubclassJSONSerializer, get_full_class_name, get_type_from_string, conclusion_to_json, is_iterable, \
    build_user_input_from_conclusion, encapsulate_user_input


class VariableVisitor(ast.NodeVisitor):
    """
    A visitor to extract all variables and comparisons from a python expression represented as an AST tree.
    """
    compares: List[Tuple[Union[ast.Name, ast.Call], ast.cmpop, Union[ast.Name, ast.Call]]]
    variables: Set[str]
    all: List[ast.BoolOp]

    def __init__(self):
        self.variables = set()
        self.attributes: Dict[ast.Name, ast.Attribute] = {}
        self.types = set()
        self.callables = set()
        self.compares = list()
        self.binary_ops = list()
        self.all = list()

    def visit_Constant(self, node):
        self.all.append(node)
        self.types.add(node)
        self.generic_visit(node)

    def visit_Call(self, node):
        self.all.append(node)
        self.callables.add(node)
        self.generic_visit(node)

    def visit_Attribute(self, node):
        self.all.append(node)
        self.attributes[node.value] = node
        self.generic_visit(node)

    def visit_BinOp(self, node):
        self.binary_ops.append(node)
        self.all.append(node)
        self.generic_visit(node)

    def visit_BoolOp(self, node):
        self.all.append(node)
        self.generic_visit(node)

    def visit_Compare(self, node):
        self.all.append(node)
        self.compares.append([node.left, node.ops[0], node.comparators[0]])
        self.generic_visit(node)

    def visit_Name(self, node):
        if f"__{node.id}__" not in dir(__builtins__) and node not in self.attributes:
            self.variables.add(node.id)
        self.generic_visit(node)


def get_used_scope(code_str, scope):
    # Parse the code into an AST
    mode = 'exec' if code_str.startswith('def') else 'eval'
    tree = ast.parse(code_str, mode=mode)

    # Walk the AST to collect used variable names
    class NameCollector(ast.NodeVisitor):
        def __init__(self):
            self.names = set()

        def visit_Name(self, node):
            if isinstance(node.ctx, ast.Load):  # We care only about variables being read
                self.names.add(node.id)

    collector = NameCollector()
    collector.visit(tree)

    # Filter the scope to include only used names
    used_scope = {k: scope[k] for k in collector.names if k in scope}
    return used_scope


class CallableExpression(SubclassJSONSerializer):
    """
    A callable that is constructed from a string statement written by an expert.
    """
    encapsulating_function: str = "def _get_value(case):"

    def __init__(self, user_input: Optional[str] = None, conclusion_type: Optional[Tuple[Type]] = None,
                 expression_tree: Optional[AST] = None,
                 scope: Optional[Dict[str, Any]] = None, conclusion: Optional[Any] = None):
        """
        Create a callable expression.

        :param user_input: The input given by the expert.
        :param conclusion_type: The type of the output of the callable.
        :param expression_tree: The AST tree parsed from the user input.
        :param scope: The scope to use for the callable expression.
        :param conclusion: The conclusion to use for the callable expression.
        """
        if user_input is None and conclusion is None:
            raise ValueError("Either user_input or conclusion must be provided.")
        if user_input is None:
            user_input = build_user_input_from_conclusion(conclusion)
        self.conclusion: Optional[Any] = conclusion
        self.user_input: str = encapsulate_user_input(user_input, self.encapsulating_function)
        if conclusion_type is not None:
            if is_iterable(conclusion_type):
                conclusion_type = tuple(conclusion_type)
            else:
                conclusion_type = (conclusion_type,)
        self.conclusion_type = conclusion_type
        self.scope: Optional[Dict[str, Any]] = scope if scope is not None else {}
        self.scope = get_used_scope(self.user_input, self.scope)
        self.expression_tree: AST = expression_tree if expression_tree else parse_string_to_expression(self.user_input)
        self.code = compile_expression_to_code(self.expression_tree)
        self.visitor = VariableVisitor()
        self.visitor.visit(self.expression_tree)

    def __call__(self, case: Any, **kwargs) -> Any:
        try:
            if self.user_input is not None:
                if not isinstance(case, Case):
                    case = create_case(case, max_recursion_idx=3)
                scope = {'case': case, **self.scope}
                output = eval(self.code, scope)
                if output is None:
                    output = scope['_get_value'](case)
                if self.conclusion_type is not None:
                    if is_iterable(output) and not isinstance(output, self.conclusion_type):
                        assert isinstance(list(output)[0], self.conclusion_type), (f"Expected output type {self.conclusion_type},"
                                                                                 f" got {type(output)}")
                    else:
                        assert isinstance(output, self.conclusion_type), (f"Expected output type {self.conclusion_type},"
                                                                          f" got {type(output)}")
                return output
            else:
                return self.conclusion
        except Exception as e:
            raise ValueError(f"Error during evaluation: {e}")

    def combine_with(self, other: 'CallableExpression') -> 'CallableExpression':
        """
        Combine this callable expression with another callable expression using the 'and' operator.
        """
        cond1_user_input = self.user_input.replace(self.encapsulating_function, "def _cond1(case):")
        cond2_user_input = other.user_input.replace(self.encapsulating_function, "def _cond2(case):")
        new_user_input = (f"{cond1_user_input}\n"
                          f"{cond2_user_input}\n"
                          f"return _cond1(case) and _cond2(case)")
        return CallableExpression(new_user_input, conclusion_type=self.conclusion_type)

    def __eq__(self, other):
        """
        Check if two callable expressions are equal.
        """
        if not isinstance(other, CallableExpression):
            return False
        return self.user_input == other.user_input and self.conclusion == other.conclusion

    def __hash__(self):
        """
        Hash the callable expression.
        """
        conclusion_hash = self.conclusion if not isinstance(self.conclusion, set) else frozenset(self.conclusion)
        return hash((self.user_input, conclusion_hash))

    def __str__(self):
        """
        Return the user string where each compare is written in a line using compare column offset start and end.
        """
        if self.user_input is None:
            return str(self.conclusion)
        binary_ops = sorted(self.visitor.binary_ops, key=lambda x: x.end_col_offset)
        binary_ops_indices = [b.end_col_offset for b in binary_ops]
        all_binary_ops = []
        prev_e = 0
        for i, e in enumerate(binary_ops_indices):
            if i == 0:
                all_binary_ops.append(self.user_input[:e])
            else:
                all_binary_ops.append(self.user_input[prev_e:e])
            prev_e = e
        return "\n".join(all_binary_ops) if len(all_binary_ops) > 0 else self.user_input

    def _to_json(self) -> Dict[str, Any]:
        return {"user_input": self.user_input,
                "conclusion_type": [get_full_class_name(t) for t in self.conclusion_type]
                if self.conclusion_type is not None else None,
                "scope": {k: get_full_class_name(v) for k, v in self.scope.items()
                          if hasattr(v, '__module__') and hasattr(v, '__name__')},
                "conclusion": conclusion_to_json(self.conclusion),
                }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> CallableExpression:
        return cls(user_input=data["user_input"],
                   conclusion_type=tuple(get_type_from_string(t) for t in data["conclusion_type"])
                   if data["conclusion_type"] else None,
                   scope={k: get_type_from_string(v) for k, v in data["scope"].items()},
                   conclusion=SubclassJSONSerializer.from_json(data["conclusion"]))


def compile_expression_to_code(expression_tree: AST) -> Any:
    """
    Compile an expression tree that was parsed from string into code that can be executed using 'eval(code)'

    :param expression_tree: The parsed expression tree.
    :return: The code that was compiled from the expression tree.
    """
    mode = 'exec' if isinstance(expression_tree, ast.Module) else 'eval'
    return compile(expression_tree, filename="<string>", mode=mode)


def parse_string_to_expression(expression_str: str) -> AST:
    """
    Parse a string statement into an AST expression.

    :param expression_str: The string which will be parsed.
    :return: The parsed expression.
    """
    mode = 'exec' if expression_str.startswith('def') else 'eval'
    tree = ast.parse(expression_str, mode=mode)
    logging.debug(f"AST parsed successfully: {ast.dump(tree)}")
    return tree



from __future__ import annotations

import json
from abc import ABC, abstractmethod

from typing_extensions import Optional, Dict, TYPE_CHECKING, List, Type, Any

from .datastructures.case import Case, CaseAttribute
from .datastructures.callable_expression import CallableExpression
from .datastructures.enums import PromptFor
from .datastructures.dataclasses import CaseQuery
from .datastructures.case import show_current_and_corner_cases
from .prompt import prompt_user_for_expression, IPythonShell
from .utils import get_all_subclasses, make_list

if TYPE_CHECKING:
    from .rdr import Rule


class Expert(ABC):
    """
    The Abstract Expert class, all experts should inherit from this class.
    An expert is a class that can provide differentiating features and conclusions for a case when asked.
    The expert can compare a case with a corner case and provide the differentiating features and can also
    provide one or multiple conclusions for a case.
    """
    all_expert_answers: Optional[List] = None
    """
    A list of all expert answers, used for testing purposes.
    """
    use_loaded_answers: bool = False
    """
    A flag to indicate if the expert should use loaded answers or not.
    """

    def __init__(self, use_loaded_answers: bool = False, append: bool = False):
        self.all_expert_answers = []
        self.use_loaded_answers = use_loaded_answers
        self.append = append

    @abstractmethod
    def ask_for_conditions(self, case_query: CaseQuery, last_evaluated_rule: Optional[Rule] = None) \
            -> CallableExpression:
        """
        Ask the expert to provide the differentiating features between two cases or unique features for a case
        that doesn't have a corner case to compare to.

        :param case_query: The case query containing the case to classify and the required target.
        :param last_evaluated_rule: The last evaluated rule.
        :return: The differentiating features as new rule conditions.
        """
        pass

    @abstractmethod
    def ask_for_conclusion(self, case_query: CaseQuery) -> Optional[CallableExpression]:
        """
        Ask the expert to provide a relational conclusion for the case.

        :param case_query: The case query containing the case to find a conclusion for.
        :return: A callable expression that can be called with a new case as an argument.
        """


class Human(Expert):
    """
    The Human Expert class, an expert that asks the human to provide differentiating features and conclusions.
    """

    def save_answers(self, path: str):
        """
        Save the expert answers to a file.

        :param path: The path to save the answers to.
        :param append: A flag to indicate if the answers should be appended to the file or not.
        """
        if self.append:
            # read the file and append the new answers
            with open(path + '.json', "r") as f:
                all_answers = json.load(f)
                all_answers.extend(self.all_expert_answers)
            with open(path + '.json', "w") as f:
                json.dump(all_answers, f)
        else:
            with open(path + '.json', "w") as f:
                json.dump(self.all_expert_answers, f)

    def load_answers(self, path: str):
        """
        Load the expert answers from a file.

        :param path: The path to load the answers from.
        """
        with open(path + '.json', "r") as f:
            self.all_expert_answers = json.load(f)

    def ask_for_conditions(self, case_query: CaseQuery,
                           last_evaluated_rule: Optional[Rule] = None) \
            -> CallableExpression:
        if not self.use_loaded_answers:
            show_current_and_corner_cases(case_query.case, {case_query.attribute_name: case_query.target_value},
                                          last_evaluated_rule=last_evaluated_rule)
        return self._get_conditions(case_query)

    def _get_conditions(self, case_query: CaseQuery) \
            -> CallableExpression:
        """
        Ask the expert to provide the differentiating features between two cases or unique features for a case
        that doesn't have a corner case to compare to.

        :param case_query: The case query containing the case to classify.
        :return: The differentiating features as new rule conditions.
        """
        user_input = None
        if self.use_loaded_answers and len(self.all_expert_answers) == 0 and self.append:
            self.use_loaded_answers = False
        if self.use_loaded_answers:
            user_input = self.all_expert_answers.pop(0)
        if user_input:
            condition = CallableExpression(user_input, bool, scope=case_query.scope)
        else:
            user_input, condition = prompt_user_for_expression(case_query, PromptFor.Conditions)
        if not self.use_loaded_answers:
            self.all_expert_answers.append(user_input)
        case_query.conditions = condition
        return condition

    def ask_for_conclusion(self, case_query: CaseQuery) -> Optional[CallableExpression]:
        """
        Ask the expert to provide a conclusion for the case.

        :param case_query: The case query containing the case to find a conclusion for.
        :return: The conclusion for the case as a callable expression.
        """
        expression: Optional[CallableExpression] = None
        if self.use_loaded_answers and len(self.all_expert_answers) == 0 and self.append:
            self.use_loaded_answers = False
        if self.use_loaded_answers:
            expert_input = self.all_expert_answers.pop(0)
            if expert_input is not None:
                expression = CallableExpression(expert_input, case_query.attribute_type,
                                                scope=case_query.scope)
        else:
            show_current_and_corner_cases(case_query.case)
            expert_input, expression = prompt_user_for_expression(case_query, PromptFor.Conclusion)
            self.all_expert_answers.append(expert_input)
        case_query.target = expression
        return expression



import json
from dataclasses import dataclass
from unittest import TestCase

from typing_extensions import List

from ripple_down_rules.datasets import load_zoo_dataset
from ripple_down_rules.datastructures.dataclasses import CaseQuery, Case
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR, MultiClassRDR, GeneralRDR
from ripple_down_rules.utils import make_set, flatten_list, serialize_dataclass, deserialize_dataclass
from test_helpers.helpers import get_fit_mcrdr, get_fit_scrdr, get_fit_grdr


@dataclass
class Position:
    x: float
    y: float


@dataclass
class Robot:
    name: str
    position: Position


class TestJSONSerialization(TestCase):
    all_cases: List[Case]
    targets: List[str]
    cache_dir: str = "./test_results"
    expert_answers_dir: str = "./test_expert_answers"

    @classmethod
    def setUpClass(cls):
        cls.all_cases, cls.targets = load_zoo_dataset(cls.cache_dir + "/zoo_dataset.pkl")

    def test_scrdr_json_serialization(self):
        scrdr, _ = get_fit_scrdr(self.all_cases, self.targets)
        filename = f"{self.cache_dir}/scrdr.json"
        scrdr.save(filename)
        scrdr = SingleClassRDR.load(filename)
        for case, target in zip(self.all_cases, self.targets):
            cat = scrdr.classify(case)
            self.assertEqual(cat, target)

    def test_mcrdr_json_serialization(self):
        mcrdr = get_fit_mcrdr(self.all_cases, self.targets)
        filename = f"{self.cache_dir}/mcrdr.json"
        mcrdr.save(filename)
        mcrdr = MultiClassRDR.load(filename)
        for case, target in zip(self.all_cases, self.targets):
            cat = mcrdr.classify(case)
            self.assertEqual(make_set(cat), make_set(target))

    def test_grdr_json_serialization(self):
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets)
        filename = f"{self.cache_dir}/grdr.json"
        grdr.save(filename)
        grdr = GeneralRDR.load(filename)
        for case, case_targets in zip(self.all_cases[:len(all_targets)], all_targets):
            cat = grdr.classify(case)
            cat = flatten_list(cat)
            case_targets = flatten_list(case_targets)
            self.assertEqual(make_set(cat), make_set(case_targets))

    def test_serialize_dataclass(self):

        robot = Robot("Robo", Position(1.0, 2.0))

        # Serialize
        json = serialize_dataclass(robot)
        # Deserialize
        reconstructed = deserialize_dataclass(json)
        assert robot == reconstructed

